@@ -1,5 +1,5 @@
 /* GMM (C) 2018 Stephane Charette <stephanecharette@gmail.com>
- * $Id: TabCanvas.cpp 2637 2018-10-16 05:55:54Z stephane $
+ * $Id: TabCanvas.cpp 2659 2018-11-11 08:54:01Z stephane $
  */
 
 #include "GMM.hpp"
@@ -11,7 +11,13 @@ TabCanvas::TabCanvas(const std::string & name) :
 	tab_is_active(false),
 	export_image(cfg().get_bool(prefix + "_export_image")),
 	timer_frequency_in_milliseconds(cfg().get_int(prefix + "_chart_redraw_ms")),
-	comm_handle(INVALID_HANDLE_VALUE)
+	comm_handle(INVALID_HANDLE_VALUE),
+	run_sum(0.0),
+	pkg_sum(0.0),
+	run_average(0.0),
+	pkg_average(0.0),
+	run_std_dev(0.0),
+	pkg_std_dev(0.0)
 {
 	const Colour pink_background(0xe6, 0xcc, 0xf2);	// total
 	const Colour blue_background(0xc0, 0xc0, 0xff);	// local
@@ -100,8 +106,8 @@ TabCanvas::TabCanvas(const std::string & name) :
 
 	for (size_t idx=0; idx < number_of_buckets; idx++)
 	{
-		total_moisture_content[idx] = 0;
-		local_moisture_content[idx] = 0;
+		run_moisture_bucket[idx] = 0;
+		pkg_moisture_bucket[idx] = 0;
 
 		addAndMakeVisible(total_distribution[idx]);
 		addAndMakeVisible(percentages		[idx]);
@@ -124,25 +130,8 @@ TabCanvas::TabCanvas(const std::string & name) :
 	kiln_num_editor	.setText("7");
 	run_editor		.setText("026");
 	pkg_editor		.setText("118");
-#if 0
-	total_moisture_content[0] = 72;
-	total_moisture_content[1] = 155;
-	total_moisture_content[2] = 227;
-	total_moisture_content[3] = 214;
-	total_moisture_content[4] = 378;
-	total_moisture_content[5] = 355;
-	total_moisture_content[6] = 524;
-	total_moisture_content[7] = 535;
-	total_moisture_content[8] = 650;
-	total_moisture_content[9] = 675;
-	total_moisture_content[10] = 532;
-	total_moisture_content[11] = 537;
-	total_moisture_content[12] = 387;
-	total_moisture_content[13] = 377;
-	total_moisture_content[14] = 220;
-	total_moisture_content[15] = 215;
-	total_moisture_content[16] = 170;
-#endif
+
+	read_all_mc_files();
 
 	// add the chart, and start the timer to redraw it every few seconds
 	addAndMakeVisible(chart);
@@ -162,7 +151,7 @@ TabCanvas::~TabCanvas(void)
 
 	stopTimer();
 
-	waitForThreadToExit(1000); // milliseconds
+	waitForThreadToExit(10000); // milliseconds
 
 	LOG_MSG("finished destructor for " << (tab_is_active ? "active" : "inactive") << " TabCanvas object with prefix=" << prefix);
 
@@ -176,16 +165,20 @@ void TabCanvas::timerCallback(void)
 	 * configuration.
 	 *
 	 * Note that since we stop the timer, do some redrawing and calculations, and then restart the drawing, it means
-	 * the timer will significantly drift over time.
+	 * the timer will significantly drift over time.  This should not be a concern since all we're doing is updating
+	 * portions of the screen, including the chart.
 	 */
 
 	stopTimer();
 
-	periodic_update();
-
-	if (timer_frequency_in_milliseconds > 0)
+	if (threadShouldExit() == false)
 	{
-		startTimer(timer_frequency_in_milliseconds); // in milliseconds
+		periodic_update();
+
+		if (timer_frequency_in_milliseconds > 0)
+		{
+			startTimer(timer_frequency_in_milliseconds); // in milliseconds
+		}
 	}
 
 	return;
@@ -194,17 +187,33 @@ void TabCanvas::timerCallback(void)
 
 void TabCanvas::periodic_update(void)
 {
-	calculate_averages_and_std_deviations();
+	if (threadShouldExit() == false)
+	{
+		calculate_averages_and_std_deviations();
+	}
 
 	// re-populate the fields with the amounts stored in the buckets
 	for (size_t idx=0; idx < number_of_buckets; idx++)
 	{
-		total_distribution[idx].setText(String(total_moisture_content[idx]));
-		local_distribution[idx].setText(String(local_moisture_content[idx]));
+		total_distribution[idx].setText(String(run_moisture_bucket[idx]));
+		local_distribution[idx].setText(String(pkg_moisture_bucket[idx]));
 	}
 
+	if (threadShouldExit() == false)
+	{
+		total_number_of_boards_editor.setText(String(run_doubles.size()));
+		local_number_of_boards_editor.setText(String(pkg_doubles.size()));
+		total_average_moisture_editor.setText(Lox::Numbers::format(run_average, 2));
+		local_average_moisture_editor.setText(Lox::Numbers::format(pkg_average, 2));
+		total_std_dev_moisture_editor.setText(Lox::Numbers::format(run_std_dev, 2));
+		local_std_dev_moisture_editor.setText(Lox::Numbers::format(pkg_std_dev, 2));
+	}
+
+//	LOG_MSG("prefix=" << prefix << ": periodic update for RUN: number_of_boards: " << run_doubles.size() << ", average: " << run_average << ", stddev: " << run_std_dev);
+//	LOG_MSG("prefix=" << prefix << ": periodic update for PKG: number_of_boards: " << pkg_doubles.size() << ", average: " << pkg_average << ", stddev: " << pkg_std_dev);
+
 	// don't bother redrawing the chart unless we have to
-	if (tab_is_active || export_image)
+	if (threadShouldExit() == false && (tab_is_active || export_image))
 	{
 		auto rect = chart.getBounds();
 		chart.setImage(draw_chart(rect.getWidth(), rect.getHeight()));
@@ -214,80 +223,56 @@ void TabCanvas::periodic_update(void)
 }
 
 
+inline double calc_std_dev(const LDouble & ld, const double average)
+{
+	double std_dev = 0.0;
+
+	// STEP #1:  Find the average of the data set.  This is already done and provided as the variable "average".
+
+	if (ld.size() > 1)
+	{
+		double sum_of_squares = 0.0;
+		for (const double & d : ld)
+		{
+			// STEP #2:  Take each value in the data set and subtract the mean from it.
+			const double val1 = d - average;
+
+			// STEP #3:  Square each of the differences.
+			const double val2 = std::pow(val1, 2.0);
+
+			// STEP #4:  Add up all of the results from Step #3 to get the sum of squares.
+			sum_of_squares += val2;
+		}
+
+		// STEP #5:  Divide the sum of squares (found in step #4) by the number of values in the data set minus one.
+		const double val3 = sum_of_squares / static_cast<double>(ld.size() - 1);
+
+		// STEP #6:  Take the square root to get the standard deviation.
+		std_dev = std::sqrt(val3);
+	}
+
+	return std_dev;
+}
+
+
 void TabCanvas::calculate_averages_and_std_deviations()
 {
-	double average_of_total_moisture = 0.0;
-	double average_of_local_moisture = 0.0;
-	double std_dev_of_total_moisture = 0.0;
-	double std_dev_of_local_moisture = 0.0;
+	// the average is now kept in the class members run_average and pkg_average,
+	// so the only thing left to do is calculate the standard deviation
 
-	double number_of_boards_total = 0.0;
-	double number_of_boards_local = 0.0;
-	double moisture_value_total = 0.0;
-	double moisture_value_local = 0.0;
-
-	for (size_t idx = 0; idx < number_of_buckets; idx++)
+	if (threadShouldExit() == false)
 	{
-		const double moisture_percentage	= static_cast<double>(idx + percentage_of_first_bucket);
-		const size_t bucket_total			= total_moisture_content[idx];
-		const size_t bucket_local			= local_moisture_content[idx];
-		number_of_boards_total				+= bucket_total;
-		number_of_boards_local				+= bucket_local;
-		moisture_value_total				+= (bucket_total * moisture_percentage);
-		moisture_value_local				+= (bucket_local * moisture_percentage);
+		run_std_dev = calc_std_dev(run_doubles, run_average);
+		pkg_std_dev = calc_std_dev(pkg_doubles, pkg_average);
 	}
 
-	if (number_of_boards_total > 0.0)
-	{
-		average_of_total_moisture = moisture_value_total / number_of_boards_total;
-	}
-	if (number_of_boards_local > 0.0)
-	{
-		average_of_local_moisture = moisture_value_local / number_of_boards_local;
-	}
-
-	// now that we have the mean (average) we can loop through again and calculate standard deviation
-
-	double sum_for_standard_deviation_total = 0.0;
-	double sum_for_standard_deviation_local = 0.0;
-
-	for (size_t idx = 0; idx < number_of_buckets; idx++)
-	{
-		const double moisture_percentage	= static_cast<double>(idx + percentage_of_first_bucket);
-		const size_t bucket_total			= total_moisture_content[idx];
-		const size_t bucket_local			= local_moisture_content[idx];
-
-		sum_for_standard_deviation_total += bucket_total * (std::pow(moisture_percentage - average_of_total_moisture, 2));
-		sum_for_standard_deviation_local += bucket_local * (std::pow(moisture_percentage - average_of_local_moisture, 2));
-	}
-
-	if (number_of_boards_total > 0.0)
-	{
-		std_dev_of_total_moisture = std::sqrt(sum_for_standard_deviation_total / number_of_boards_total);
-	}
-	if (number_of_boards_local > 0.0)
-	{
-		std_dev_of_local_moisture = std::sqrt(sum_for_standard_deviation_local / number_of_boards_local);
-	}
-
-//	LOG_MSG("TOTAL:  number_of_boards: " << number_of_boards_total << ", moisture_value: " << moisture_value_total << ", average: " << average_of_total_moisture << ", stddev: " << std_dev_of_total_moisture);
-//	LOG_MSG("LOCAL:  number_of_boards: " << number_of_boards_local << ", moisture_value: " << moisture_value_local << ", average: " << average_of_local_moisture << ", stddev: " << std_dev_of_local_moisture);
-
-	total_number_of_boards_editor.setText(String(number_of_boards_total));
-	total_average_moisture_editor.setText(Lox::Numbers::format(average_of_total_moisture, 2).c_str());
-	total_std_dev_moisture_editor.setText(Lox::Numbers::format(std_dev_of_total_moisture, 2).c_str());
-
-	local_number_of_boards_editor.setText(String(number_of_boards_local));
-	local_average_moisture_editor.setText(Lox::Numbers::format(average_of_local_moisture, 2).c_str());
-	local_std_dev_moisture_editor.setText(Lox::Numbers::format(std_dev_of_local_moisture, 2).c_str());
-
 	return;
 }
 
 
 Image TabCanvas::draw_chart(const size_t width, const size_t height)
 {
-	if (tab_is_active == false && export_image == false)
+	if (threadShouldExit() || (tab_is_active == false && export_image == false))
 	{
 		// don't bother creating images for inactive tabs
 		return Image();
@@ -318,12 +303,12 @@ Image TabCanvas::draw_chart(const size_t width, const size_t height)
 	const float width_float				= static_cast<float>(width);
 	const float height_float			= static_cast<float>(height);
 	const float h_step					= 1.0f / static_cast<float>(number_of_buckets);
-	const float v_step					= 1.0f / 5.0f;
+	const float v_step					= 1.0f / 5.0f; // 20%, 40%, 60%, 80%, and 100%
 	const float width_of_columns		= h_step * width_float;
 	const float one_seventh_of_column	= 1.0f * width_of_columns / 7.0f;
 	const float three_seventh_of_column	= 3.0f * width_of_columns / 7.0f;
 
-	/* For drawing purposes, each column is divided into 7 parts.
+	/* For drawing purposes, each column is divided into 7 equal parts.
 	 * - 1st part is empty space.
 	 * - 2nd through 4th is the pink.
 	 * - 4th through 6th is the blue.
@@ -345,33 +330,23 @@ Image TabCanvas::draw_chart(const size_t width, const size_t height)
 		g.drawRect(x * width_float, 0.0f, 1.0f, height_float);
 	}
 
-	// determine the maximum moisture content value
-	size_t max_total_moisture_value = 0;
-	size_t max_local_moisture_value = 0;
+	// determine the maximum value in all the buckets (initially set to "1" to prevent divide-by-zero later)
+	size_t max_run_bucket_value = 1;
+	size_t max_pkg_bucket_value = 1;
 	for (size_t idx = 0; idx < number_of_buckets; idx++)
 	{
-		if (total_moisture_content[idx] > max_total_moisture_value)
-		{
-			max_total_moisture_value = total_moisture_content[idx];
-		}
-		if (local_moisture_content[idx] > max_local_moisture_value)
-		{
-			max_local_moisture_value = local_moisture_content[idx];
-		}
+		if (run_moisture_bucket[idx] > max_run_bucket_value)	max_run_bucket_value = run_moisture_bucket[idx];
+		if (pkg_moisture_bucket[idx] > max_pkg_bucket_value)	max_pkg_bucket_value = pkg_moisture_bucket[idx];
 	}
 
-	// prevent divide-by-zero
-	if (max_total_moisture_value == 0) max_total_moisture_value = 1;
-	if (max_local_moisture_value == 0) max_local_moisture_value = 1;
-
 	// draw the columns of data
-	for (size_t idx = 0; idx < number_of_buckets; idx++)
+	for (size_t idx = 0; threadShouldExit() == false && idx < number_of_buckets; idx++)
 	{
 		const float col_x = idx * h_step * width_float;
 
 		// the graph is based on % of the maximum, not on the raw numbers, so we need to convert this bucket to a percentage
-		const float percent_total = (float)total_moisture_content[idx] / (float)max_total_moisture_value;
-		const float percent_local = (float)local_moisture_content[idx] / (float)max_local_moisture_value;
+		const float percent_total = (float)run_moisture_bucket[idx] / (float)max_run_bucket_value;
+		const float percent_local = (float)pkg_moisture_bucket[idx] / (float)max_pkg_bucket_value;
 
 		// 0% means the bottom of the chart, while 100% means the top of the chart
 		// +2 offset means the bar will end 2 pixels from the top of the chart, leaving room for the black border and 1 pixel of white
@@ -401,7 +376,7 @@ Image TabCanvas::draw_chart(const size_t width, const size_t height)
 		}
 	}
 
-	// add % text label at the bottom of the image
+	// add % text labels at the bottom of the image
 	g.setColour(Colours::black);
 	const float label_h = 15;
 	const float label_y = height - label_h;
@@ -415,7 +390,7 @@ Image TabCanvas::draw_chart(const size_t width, const size_t height)
 	}
 
 	// add warnings and errors to the top-left of the image
-	if (warnings_and_errors.empty() == false)
+	if (warnings_and_errors.empty() == false && threadShouldExit() == false)
 	{
 		g.setColour(Colours::red);
 		float y = 0.0f;
@@ -426,7 +401,7 @@ Image TabCanvas::draw_chart(const size_t width, const size_t height)
 		}
 	}
 
-	if (export_image)
+	if (export_image && threadShouldExit() == false)
 	{
 		// save to a .png file for debug purposes
 		File file = File::getSpecialLocation(File::SpecialLocationType::tempDirectory).getChildFile(prefix + "_chart.png");
@@ -562,7 +537,7 @@ void TabCanvas::device_thread(void)
 	const bool run_simulation		=	cfg().get_bool	(prefix + "_run_simulation"		);
 	const std::regex rx				(	cfg().get_str	(prefix + "_regex_pattern"		));
 
-	LOG_MSG("device thread for " << device_name << " has started (name=\"" << tab_name << "\", prefix=\"" << prefix << "\")");
+	LOG_MSG(device_name << ": device thread for " << device_name << " has started (name=\"" << tab_name << "\", prefix=\"" << prefix << "\")");
 
 	for (const auto k : cfg().getAllProperties().getAllKeys())
 	{
@@ -598,32 +573,64 @@ void TabCanvas::device_thread(void)
 			str = read_from_device(read_buffer);
 		}
 
-		if (str.empty() == false)
+		if (str.empty() == false && threadShouldExit() == false)
 		{
 			std::smatch matches;
 			const bool found = std::regex_search(str, matches, rx);
 
 			if (found)
 			{
-				const double data = std::stod(matches[1]);
-				LOG_MSG(device_name << ": str=" << String(str).trim().toStdString() << ", data=" << data);
+				// There should be 2 values:  the peak, and the average.  We want to use the average when possible.
+				// If the average is <= 1.0, then use the peak instead since sometimes the sensors seem to send the
+				// value 0.0 for the average.  See email thread between Dominic and Stephane from 2018-10-18.
+				const double peak_value		= (matches.size() > 1 ? std::stod(matches[1]) : 0.0);
+				const double average_value	= (matches.size() > 2 ? std::stod(matches[2]) : 0.0);
+				double value = average_value;
+				if (value < 1.0)
+				{
+					value = peak_value;
+					LOG_MSG(device_name << ": invalid average value from sensor: str=" << String(str).trim().toStdString());
+				}
+				else
+				{
+					LOG_MSG(device_name << ": str=" << String(str).trim().toStdString() << ", reading=" << value);
+				}
 
-				// determine which bucket this should fit into
-				int idx = static_cast<int>(data) - percentage_of_first_bucket;
-				if (idx < 0)					idx = 0;
-				if (idx >= number_of_buckets)	idx = number_of_buckets - 1;
+				if (value > 0.0)
+				{
+					// determine which bucket this should fit into
+					int idx = static_cast<int>(value) - percentage_of_first_bucket;
+					if (idx < 0)					idx = 0;
+					if (idx >= number_of_buckets)	idx = number_of_buckets - 1;
 
-				local_moisture_content[idx] ++;
-				total_moisture_content[idx] ++;
-				LOG_MSG(device_name << ": data=" << data << ", incrementing bucket #" << idx << " to " << local_moisture_content[idx]);
+					run_doubles.push_back(value);
+					pkg_doubles.push_back(value);
+
+					run_sum += value;
+					pkg_sum += value;
+
+					run_average = run_sum / static_cast<double>(run_doubles.size());
+					pkg_average = pkg_sum / static_cast<double>(pkg_doubles.size());
+
+					run_moisture_bucket[idx] ++;
+					pkg_moisture_bucket[idx] ++;
+					LOG_MSG(device_name << ": value=" << value << ", incrementing bucket #" << idx << " to " << pkg_moisture_bucket[idx]);
+
+					// every once in a while, write what we have back out to disk so we can restore things when we come back up
+					if (0 == pkg_doubles.size() % 100)
+					{
+						write_mc_file();
+					}
+				}
 			}
 			else
 			{
 				std::string msg;
 				for (const char c : str)
 				{
-					if (c > 32 && c < 127)	msg += c;
-					else					msg += "<" + std::to_string((int)c) + ">";
+					const int i = static_cast<int>(c);
+					if (i > 32 && i < 127)	msg += c; // normal ASCII character
+					else					msg += "<" + std::to_string(i) + ">"; // non-ASCII printable character
 				}
 				LOG_MSG(device_name << ": failed to parse string: len=" << str.size() << ": \"" << msg << "\"");
 			}
@@ -636,11 +643,14 @@ void TabCanvas::device_thread(void)
 		{
 			// this is a problem -- either the time interval is too small, or the
 			// computer/VM/whatever where we are running is too slow to keep up
-			//
+			const auto int_ms = std::chrono::duration_cast<std::chrono::milliseconds>(time_now - next_read);
+
+			LOG_MSG(device_name << ": warning: time overflow by " << int_ms.count() << " milliseconds");
+
 			// reset the next read time as if it was now, and hope this doesn't happen again
 			next_read = time_now;
 		}
-		else
+		else if (threadShouldExit() == false)
 		{
 //			LOG_MSG(device_name << ": sleeping for " << std::chrono::duration_cast<std::chrono::milliseconds>(next_read - time_now).count() << " milliseconds");
 			std::this_thread::sleep_until(next_read);
@@ -648,7 +658,111 @@ void TabCanvas::device_thread(void)
 		next_read += interval;
 	}
 
-	LOG_MSG("device thread for " << device_name << " is exiting (prefix=" << prefix << ")");
+	LOG_MSG("prefix=" << prefix << ": detected shutdown signal");
+
+	write_mc_file();
+
+	LOG_MSG("prefix=" << prefix << ": device thread for \"" << tab_name << "\" on " << device_name << " is exiting");
+
+	return;
+}
+
+
+void TabCanvas::write_mc_file(void)
+{
+	// make sure to see the Doxygen documentation to learn about all the fields
+
+	String dirname = cfg().get_str("output_directory");
+	dirname = dirname.replaceFirstOccurrenceOf("%Y", Time::getCurrentTime().formatted("%Y"));
+
+	File dir(dirname);
+	dir.createDirectory();
+
+	const std::string filename = dir.getChildFile(prefix + "_test.mc").getFullPathName().toStdString();
+	LOG_MSG("prefix=" << prefix << ": writing data out to: " << filename << " (pkg size=" << pkg_doubles.size() << ", pkg avg=" <<  pkg_average << ")");
+
+	// LINE #1:
+
+	std::ofstream fs(filename);
+	fs << std::fixed << std::setw(4) << std::setfill('0') << pkg_doubles.size() << ",";
+
+	for (size_t idx = 0; idx < number_of_buckets - 6; idx++)
+	{
+		fs << std::setw(4) << std::setfill('0') << pkg_moisture_bucket[idx] << ",";
+	}
+
+	fs << "000,000,";	// unknown fields
+
+	fs << pkg_std_dev << "," << pkg_average << ",000" << std::endl;
+
+	// LINE #2:
+
+	for (const double d : pkg_doubles)
+	{
+		fs << std::setw(4) << std::setfill('0') << std::setprecision(1) << d << ",";
+	}
+	fs << "." << std::endl;
+
+	LOG_MSG("prefix=" << prefix << ": done writing data to " << filename);
+
+	return;
+}
+
+
+void TabCanvas::read_all_mc_files(void)
+{
+	String dirname = cfg().get_str("output_directory");
+	dirname = dirname.replaceFirstOccurrenceOf("%Y", Time::getCurrentTime().formatted("%Y"));
+
+	File dir(dirname);
+	dir.createDirectory();
+
+	const std::string filename = dir.getChildFile(prefix + "_test.mc").getFullPathName().toStdString();
+
+	std::ifstream fs(filename);
+	std::string str;
+	std::getline(fs, str); // ignore the first line
+	std::getline(fs, str); // parse the 2nd line
+	fs.close();
+
+//	LOG_MSG("prefix=" << prefix << ": filename=" << filename << ": 2nd line to parse: \"" << str << "\"");
+
+	// replace comma with spaces to make it easier to parse the individual fields
+	while (true)
+	{
+		const size_t pos = str.find(',');
+		if (pos == std::string::npos)
+		{
+			break;
+		}
+		str[pos] = ' ';
+	}
+
+	std::stringstream ss(str);
+	while (ss)
+	{
+		ss >> str;
+		if (str.empty() == false && str != ".")
+		{
+			const double d = std::stod(str);
+			if (d > 0.0)
+			{
+				int idx = static_cast<int>(d) - percentage_of_first_bucket;
+				if (idx < 0)					idx = 0;
+				if (idx >= number_of_buckets)	idx = number_of_buckets - 1;
+
+				run_moisture_bucket[idx] ++;
+				run_doubles.push_back(d);
+				run_sum += d;
+			}
+		}
+	}
+
+	if (run_doubles.empty() == false)
+	{
+		run_average = run_sum / static_cast<double>(run_doubles.size());
+	}
+	LOG_MSG("prefix=" << prefix << ": filename=" << filename << ": imported " << run_doubles.size() << " existing entries, average is " << run_average);
 
 	return;
 }
